import os, time, shutil


"""
TODO:
    - Add dunder functions
    - Compress data when weights bigger than x
"""

class Checkpoint:
    """
        Keep track of model iterations by saving checkpoints.
        Generate checkpoints by saving model weights.

        Generates file names of form: <filename>_<timestamp

        Parameters:
            path (str): The path where the checkpoints getting saved.
            filename (str): The filename prefix to use for checkpoints.
    """

    def __init__(self, path, filename="model", extension=None, sub_dir=True):

        self.PATH = path if not sub_dir else os.path.join(path, filename)
        self.FILENAME = filename
        self.EXTENSION = extension if not (extension is None) else "pb"
        self.checkpoints = []


    def __create_checkpoint_dir(self):
        if not os.path.isdir(os.path.join(self.PATH)):
            os.mkdir(self.PATH)


    def new(self, model):
        """
            Create a new checkpoint, saving the model parameters.

            Parameters:
                model (tf.Model): The model for which parameters to save.
        """
        self.__create_checkpoint_dir()

        # Build checkpoint path
        timestamp = time.strftime("%d_%m_%y_%H_%M_%S")
        full_file_name = self.FILENAME + "_" + timestamp + "." + self.EXTENSION
        self.checkpoints.append(full_file_name)

        # Create the new checkpoint
        checkpoint_path = os.path.join(self.PATH, full_file_name)
        model.save_weights(checkpoint_path)

    
    def __try_checkpoints_recovery(self):
        # Try recover checkpoints from directory
        checkpoint_dir = os.listdir(self.PATH)
        for file in checkpoint_dir:
            
            if file.endswith(self.EXTENSION):
                self.checkpoints.append(file)
        
        # Still no checkpoints after recovery attempt
        if len(self.checkpoints) == 0:
            raise ValueError("Checkpoint list is empty. Recovery attempt failed. Check if there are any files within the checkpoint directory {} with extension .{} .".format(self.PATH, self.EXTENSION))


    def load(self, model, iteration=None):
        """
            Load the given checkpoint weights into the model.
            **Deprecated**: Use get path to get the path inside a specific bayes mdoel.


            Parameters:
                model (tf.Model): The tensorflow model to load the weights into
                iteration (int): The checkpoint to load 
        """
        checkpoint_path = self.path(iteration=iteration)
        model.load_weights(checkpoint_path)


    def path(self, iteration=None):
        """
            Get the path for nth checkpoint.

            Parameters:
                iteration (int): The path of given n-th checkpoint to retrieve. (default: last checkpoint)

            Returns:
                (str) The path to the checkpoint at given iteration
        """
        if len(self.checkpoints) == 0:
            # TODO: Checkpoint recovery can fail when path exists but model not of same arch.
            self.__try_checkpoints_recovery()

        checkpoint_name = None
        if iteration is None:
            # Select the last checkpoint
            checkpoint_name = self.checkpoints[-1]

        elif iteration < 0:
            # Include -1?
            raise ValueError("Can't load negative iteration {}. Only positive values for iteration alloweds".format(iteration)) 

        elif len(iteration) > iteration:
            checkpoint_name = self.checkpoints[iteration]

        else:
            # Iteration out of range
            raise ValueError("Can't load iteration {}. There's only {} checkpoints.".format(itreation, len(self.checkpoints)))

        return os.path.join(self.PATH, checkpoint_name)
        

    def clean(self):
        """
            Remove all checkpoint files.
        """
        for self.FILENAME in self.checkpoints:
            shutil.rmtree(self.PATH)


    def empty(self):
        """
            Are there any checkpoints?
        """
        
        if self.__len__() == 0:
            return True
        
        return False


    # ----------------------
    # Dunder functions
    # ---------------------------
    def __getitem__(self, key):
        return self.checkpoints[key]

    
    def __len__(self):
        return len(self.checkpoints)



from .experiment_metrics import ExperimentSuitMetrics
from .metrics import Metrics, aggregates_per_key, save_history

from .oracle import Oracle
from .acquisition_function import AcquisitionFunction

from .config import Config
from .pool import Pool

from .dataset import Dataset
from .active_learning_loop import ActiveLearningLoop
from .experiment_suit import ExperimentSuit


import math
import numpy as np

import tensorflow as tf
from .utils import setup_logger


class AcquisitionFunction:
    """
        Query a model for next datapoints that should be labeled.
        Already implemented bayesian models implement the following acquisition
        functions:

        - Max Entropy (\"max_entropy\")
        - Bald (\"bald\")
        - Variation Ratios (\"max_var_ratio\")
        - Mean standard deviation (\"mean_std\")
        - Randomized selection (\"random\")

        Parameters:
            fn_name (str): The name of the acquisition function to use.
            batch_size (int): The number of batches to split the data into for processing. (default=None)
            verbose (bool): Apply debugging? (default=False)

        Attributes:
            name (str): The name of the acquisition function to apply.
            fn (function): The acquisition function to execute.
            batch_size (int): to configure the processing in batches (default=None)
    """

    def __init__(self, fn_name, batch_size=None, verbose=False, **kwargs):
        self.logger = setup_logger(verbose, "Acquisition Function Logger")

        self.name = fn_name
        self.fn = None
        self.batch_size = batch_size

        self.kwargs = kwargs

        # Set passed kwargs as additional attributes, not overwriting existing
        # for key, value in kwargs.items():
        #     if not hasattr(self, key):
        #         setattr(self, key, value)


    def __call__(self, model, pool, step_size=20, **kwargs):
        """
            
            Parameter:
                model (Model): The model to use for the computation of acquistion functions.
                pool (Pool): The pool of unlabeled data.
                step_size (int): Number of datapoints to collect for next active learning iteration.

            Returns:
                (numpy.ndarray) Indices
        """
        self.logger.info("Step-size: {}".format(step_size))
        self.logger.info("Function: {}".format(self.name))
        self.logger.info("Kwargs: {}".format(kwargs))

        # Set initial acquistion function
        if self.fn is None:
            self.fn = self._set_fn(model)

        data, indices = pool.get_unlabeled_data()
        # data = pool.get_data()
        # indices = pool.get_indices()

        # No need for batch processing, random selection of datapoints.
        if self.name == "random":
            return self.fn(indices, data, step_size=step_size, **kwargs)            

        # Iterate throug batches of data
        results = None
        num_datapoints = len(data)
        self.logger.info("Unlabeled Pool size: {}".format(num_datapoints))
           
        # Single batch processing
        if self.batch_size is None:
            self.batch_size = len(data)

        # Multi batch Processing
        num_batches = math.ceil(num_datapoints/self.batch_size)
        batches = np.array_split(data, num_batches, axis=0)
        results = []
        for batch in batches:
            
            input_tensor = tf.convert_to_tensor(batch)
            output_tensor = self.fn(input_tensor, **kwargs)
            results.append(output_tensor)
        
        stacked = np.hstack(results)
        num_of_elements_to_select = self._adapt_selection_num(len(stacked), step_size)
        return self.__select_first(stacked, indices, num_of_elements_to_select)


    def _adapt_selection_num(self, num_indices, num_to_select):
        """
            Check if n datapoints are available at all, else adapt the number of datapoints to select.

            Parameters:
                num_indices (int): The amount of total indices to select from
                num_to_select (int): The number of datapoints to select

            Returns:
                (int) the adapted number datapoints to select from a collection.
        """
        if num_indices == 0:
            raise ArgumentError("Can't select {} datapoints, all data is labeled.".format(num_to_select))

        if num_indices < num_to_select:
            return num_indices
        
        return num_to_select


    def _set_fn(self, model):
        """
            Set the function to use for acquisition.
            
            Parameters:
                name (str): The name of the acquisition to use.

            Returns:
                (function): The function to use for acquisition.
        """

        query_fn = model.get_query_fn(self.name)
        if query_fn is None:
            self.logger.debug("Set acquisition function: random baseline.")
            self.name = "random"
            return self._random

        else:
            return query_fn
    

    def _random(self, indices, data, step_size=5, **kwargs):
        """
            Randomly select a number of datapoints from the dataset.
            Baseline for comparison purposes.

            FIX: Random selection is wrong!!!

            Parameters:
                model (BayesianModel): The model to perform active learning on.
                pool (DataPool): The pool of data to use.
                num (int): Numbers of indices to draw from unlabeled data.
           
            Returns:
                (numpy.ndarray): Randomly selected indices for next training.
        """
        available_indices = np.linspace(0, len(data)-1, len(data), dtype=int)
        step_size = self._adapt_selection_num(len(available_indices), step_size)
        selected = np.random.choice(available_indices, step_size, replace=False).astype(int)
        return indices[selected], data[selected]


    def __select_first(self, predictions, indices, n):
        """
            Select n biggest elements from k- predictions.

            Parameters:
                predictions (numpy.ndarray): The predictions made by the network

            Returns: 
                (numpy.ndarray) indices of n-biggest predictions.
        """
        
        sorted_keys = np.argsort(predictions)
        n_biggest_keys = sorted_keys[-n:]
        return indices[n_biggest_keys], predictions[n_biggest_keys]


    # ----------
    # Dunder
    # ------------------

    def __str__(self):
        return self.name


    # ----------
    # Getter-/Setter
    # ------------------
    
    def set_fn(self, fn):
        self.fn = fn

    def get_name(self):
        return self.name


import tensorflow as tf
import numpy as np
from tqdm import tqdm


def disable_batch_norm(model):
    """
        Disable batch normalization for activation of dropout during prediction.

        Parameters:
            - model (tf.Model) Tensorflow neural network model.
    """
    
    disabled = False
    for l in model.layers:
        if l.__class__.__name__ == "BatchNormalization":
            disabled = True
            l.trainable = False

    if disabled:
        print("Disabled BatchNorm-Layers.")


def predict_n(model, point, n_times=5, enable_tqdm=True):
    """
        Helper to perform n-predictions for given bayesian model.
        Returns predictions in numpy array.

        Parameters:
            - model (tf.Model | torch.Model) The model to be used for the prediction
            - point (np.ndarray) The datapoint to be used for prediction
            - n_times (int) The number of predictions to be made
            - enabel_tqdm (bool) Should tqdm be used for progress indication?

        TODO: 
            - [ ] extend to be useable with pytorch aswell
            - [ ] Test if first dimension of output can always be omitted
        
        Returns:
            - (np.ndarray)  of shape (ntimes, ...model output shape)
    """

    if n_times < 1 or not int(n_times):
        raise ArgumentError("At least single prediction needed. But for n_times {} given.".format(n_times))

    # Prepare layers of bayesian model for prediction
    disable_batch_norm(model)

    # Make n-predictions
    output = None
    iterator = tqdm(range(n_times)) if enable_tqdm else range(n_times)
    for i in iterator:

        # Set initial shape
        result = model(point, training=True)
        if output is None:
            output = np.zeros(tuple([n_times]+list(result.shape[1:])))

        output[i] = result

    tf.keras.backend.clear_session()
    return output


def batch_predict_n(model, data, n_times=5, enable_tqdm=True):
    """
        Perform n-predictions per data point in colletion of data points.

        Parameters:
            - model (tf.Model | torch.Model) The model to perform predictions on
            - data (np.ndarray) The data to perform the predictions on
            - n_times (int) total number of predictions per input point
            - fun_pred (function) 

        Returns:
            - (np.ndarray) of shape (num data points, n_times, ...model output shape)
    """

    disable_batch_norm(model)
    data_points = data.shape[0]
    output = None
    # batch_size = data_points if batch_size is None else batch_size
    iterator = tqdm(range(n_times)) if enable_tqdm else range(n_times)
    for i in iterator:

        # Set initial shape
        result = model(data, training=True)
        if output is None:
            output = np.zeros(tuple([n_times] + list(result.shape)))

        output[i] = result

    tf.keras.backend.clear_session()
    return output.reshape(tuple([data_points, n_times] + list(output.shape[2:])))



def measures(predictions, batch_dim=False):
    """
        Helper to calculate regular measures for bayesian neural networks.

        Parameters:
            - predictions (np.ndarray) Array with n-predictions.
            - batch_dim (bool) Existing batch dimension at first index.

        Returns:
            - (np.ndarray, np.ndarray) The mean and std of predictions
    """

    axis = 0
    if batch_dim:
        axis = 1

    avg = np.mean(predictions, axis=axis)
    var = np.var(predictions, axis=axis)

    return (avg, var)


def uncertainty():
    """
        Helper to calculate specific uncertainty measures.
    """
    pass




import tensorflow as tf
import tensorflow.keras as keras

from .dataset import Dataset


class KerasDataset(Dataset):
    """
        Load standard datasets of tensorflow/keras, split and 

        Supported names:
            - mnist
            - cifar10
            - cifar100
            - reuters
            - imdb
            - fashion_mnist
            - boston_housing

        Parameters:
            name (str): The name of the keras dataset to load.
            init_size (int): Initial labeled pool size.
            init_indices (list|numpy.ndarray): List or array of initial labeled pool indices.
            transform_inputs (function): Transformation applied on train and test inputs (x_train, x_test).
            transform_outputs (function): Transformation applied on train and test targets (y_train, y_test).
            **kwargs: Additional parameters, passed on to the load_data function
    """

    def __init__(self, name, init_size=0, init_indices=None, transform_inputs=None, transform_outputs=None, **kwargs):
        (x_train, y_train), (x_test, y_test) = self.__load_dataset(name, **kwargs)

        x_train = self.__transform_data(transform_inputs, x_train)
        y_train = self.__transform_data(transform_outputs, y_train)

        x_test = self.__transform_data(transform_inputs, x_test)
        y_test = self.__transform_data(transform_outputs, y_test)

        super().__init__(
            x_train, 
            y_train,
            test=(x_test, y_test),
            init_size=init_size,
            init_indices=init_indices
        )

    
    def __transform_data(self, fn_transform, data):
        """
            Apply transformation on the dataset.
        """
        if fn_transform is None:
            return

        return fn_transform(data)


    def __load_dataset(self, name, **kwargs):
        """
            Load a standard keras dataset by it's name. A list of datasets can be found [here](https://keras.io/api/datasets/).

            Supported names:
            - mnist
            - cifar10
            - cifar100
            - reuters
            - imdb
            - fashion_mnist
            - boston_housing

            Parameters:
                name (str): The name of the dataset
                **kwargs (dict): Dataset specific parameters. Depending on the dataset loaded.
        """

        name = name.lower()
        if name == "mnist":
            path = kwargs.get("path", "mnist.npz")
            return keras.datasets.mnist.load_data(path=path)

        elif name == "cifar10":
            return keras.datasets.cifar10.load_data()
        
        elif name == "cifar100":
            label_mode = kwargs.get("label_mode", "fine")
            return keras.datasets.cifar100.load_data(label_mode=label_mode)

        elif name == "imdb":
            return keras.datasets.imdb.load_data(**kwargs)

        elif name == "reuters":
            return keras.datasets.reuters.load_data(**kwargs)

        elif name == "fashion_mnist":
            return keras.datasets.fashion_mnist.load_data()

        elif name == "boston_housing": 
            params = self.__filter_kwargs(["path", "test_split", "seed"], kwargs)
            return keras.datasets.boston_housing.load_data(**params)

        raise ValueError("Error in StandardDataset.__init__(). Can't find keras dataset for name {}.".format(name))


    # -------------------
    # Utilities
    # -------------------------
    
    def __filter_kwargs(self, names, kwargs):
        """
            Collect 
        """

        parsed = {}
        for key, value in kwargs.items():

            if key in names:
                parsed[key] = value

        return parsed


    def __collect_and_set_defaults(self, names_and_defaults, **kwargs):
        """
            Deprecated

            Parameters:
                names_and_defaults (dict): A dictionary of parameter names and defaults.
                **kwargs (dict): Parameters received from user.

            Returns:
                (dict) with loaded parameters, while eventually setting default params.
        """

        parsed = {}
        for name, default in names_and_defaults.items():
            parsed[name] = kwargs.get(name, default)

            if name in kwargs:
                del kwargs[name]
            
        
        return parsed, kwargs


from enum import Enum
import numpy as np


class LabelType(Enum):
    CLASS_LABEL=1,
    ONE_HOT_VECTOR=2,
    VALUE=3

class OracleMode(Enum):
    PSEUDO=1,
    ANNOTATE=2


class Oracle:
    """
        Oracle handles the labeling process for input values.
        
        Parameters:
            callback (Callback): Function to call for user input for input values. Function receives (pool, indices)
            pseudo_mode (bool): Active learning environment in pseudo mode?
    """

    def __init__(self, callback=None, pseudo_mode=False):
        self.__annotation_callback = callback
        self.pseudo_mode = pseudo_mode


    def init(self, pool, size, pseudo_mode=None):
        """
            Initialize pool with given number of samples.

            Parameters:
                pool (Pool): holding information about already labeled targets.
                size (int): number of elements to initialize the pool with.
                pseudo_mode (bool): Whether or not pseudo labeling of inputs. (Only applicable when pool initialized with targets)
        """

        oracle_in_pseudo_mode = self.is_pseudo(pseudo_mode) 

        if self.is_pseudo(pseudo_mode) and pool.is_pseudo():
            pool.init(size)
            return
        

        if self.__annotation_callback is None:
            raise ValueError("Error in Oracle.init(). Can't initialize pool because no callback function was set.")

        unlabeled_indices = pool.get_unlabeled_indices()
        indices = np.random.choice(unlabeled_indices, size, replace=False)
        self.__annotation_callback(pool, indices)


    def annotate(self, pool, indices, pseudo_mode=None):
        """
            Create annotations for given indices and update the pool.

            Parameters:
                pool (Pool): The pool holding information about already annotated inputs.
                indices (numpy.ndarray|list(int)): Indices indicating which inputs to annotate.
        """
        
        # Pseudo mode, use already known labels
        if pool.is_pseudo() and self.is_pseudo(pseudo_mode): 
            pool.annotate(indices)
            return 

        if self.__annotation_callback is None:
            raise ValueError("Error in Oracle.annotate(). Oracle not in pseudo-mode and callback is None.")

        self.__annotation_callback(pool, indices)

    
    def is_pseudo(self, mode=None):
        """
            Is the oracle put into pseudo labeling mode?
            Meaning: when pool is also in pseudo mode, labels will automatically be set by using known labels.
        """

        if mode is not None:
            if not isinstance(mode, bool):
                raise ValueError("Error in Oracle. Value of pseudo_mode is not boolean.")

            return mode
            
        else:
            return self.pseudo_mode
        



        





def log_mem_usage(logger):
    import nvidia_smi
    import psutil
    from psutil._common import bytes2human

    nvidia_smi.nvmlInit()
    handle = nvidia_smi.nvmlDeviceGetHandleByIndex(0)
    info = nvidia_smi.nvmlDeviceGetMemoryInfo(handle)

    virtual_mem = psutil.virtual_memory()
    logger.info("(total) RAM: {}".format(bytes2human(virtual_mem.total)))
    logger.info("(free) RAM: {}".format(bytes2human(virtual_mem.free)))
    logger.info("(used) RAM: {}".format(bytes2human(virtual_mem.used)))

    logger.info("GPU----------")
    logger.info("Total GPU: {}".format(bytes2human(info.total)))
    logger.info("Free GPU: {}".format(bytes2human(info.free)))
    logger.info("Used GPU: {}".format(bytes2human(info.used)))
    nvidia_smi.nvmlShutdown()


import numpy as np
from . import Model, ModelType, Mode



class SWAG(Model):

    # https://github.com/DD2412-Final-Projects/swag-reproduced/blob/master/train_swag.py

    def __init__(self, model):
        super().__init__(model, config, model_type=ModelType.SWAG, **kwargs)

        self.second_moment = None
        self.cov_mat = None

    
    def __call__(self, inputs, **kwargs):
        pass

    
    def evaluate(self, inputs, targets, **kwargs):
        pass

    
    def fit(self, inputs, targets, **kwargs):
        pass


    def get_query_fn(self, name):
    
        # if name == "max_entropy":
        #     return self.__max_entropy
        
        # if name == "bald":
        #     return self.__bald
        
        # if name == "max_var_ratio":
        #     return self.__max_var_ratio

        # if name == "std_mean":
        #     return self.__std_mean

        return None


import os, sys, csv

class Metrics:
    """
        Uses the given path to create 
        Prepares and writes metrics into a csv file.

        Parameters:
            base_path (str): The base path where to save the metrics.
            keys (list(str)): A list of keys.
    """

    def __init__(self, base_path, keys=["accuracy", "loss"]):
        self.metric_keys = keys
        self.BASE_PATH = base_path
        self.EXT = "csv"

        # CSV Parameters
        self.delimiter = " "
        self.quotechar = "\""
        self.quoting = csv.QUOTE_MINIMAL


    def write_line(self, filename, values):
        file_path = os.path.join(self.BASE_PATH, filename+"."+self.EXT)
        with open(file_path, "a", newline="") as csv_file:
            pass
    

    def write(self, filename, values):
        """
            Write given values into a csv file.

            Parameters:
                filename (str): The name of the file.
                values (list(dict)): A dictionary of metrics/values to write into a .csv file.
        """

        file_path = os.path.join(self.BASE_PATH, filename+"."+self.EXT)
        with open(file_path, "w", newline="") as csv_file:
            
            # Setup csv file
            file_writer = csv.DictWriter(
                csv_file, delimiter=self.delimiter, 
                quotechar=self.quotechar, quoting=self.quoting, fieldnames=self.metric_keys)

            # Create content of csv file
            file_writer.writeheader()
            for line in values:
                collected = self.collect(line)
                file_writer.writerow(collected)

    
    def read(self, filename):
        """
            Read a .csv file of metrics.

            Parameters:
                filename (str): The filename to read in.

            Returns:
                (list(dict)) a list of metric values, per trained iteration.
        """

        values = []

        if not ("."+self.EXT in filename):
            filename = filename + "." + self.EXT 

        file_path = os.path.join(self.BASE_PATH, filename)
        with open(file_path, "r") as csv_file:

            reader = csv.DictReader(
                filter(lambda row: row[0] != "#", csv_file), 
                delimiter=self.delimiter, 
                quotechar=self.quotechar
            )

            for row in reader:
                values.append(row)

        return values



    # -------------
    # Utilities
    # ------------------

    def collect(self, values, keys=None):
        """
            Collect metric values from a dictionary of values.

            Parameter:
                values (dict): A collection of values collected during training

            Returns:
                (dict) A subset of metrics extracted from the values. 
        """
        # Set default keys to use
        if keys is None:
            keys = self.metric_keys

        return {key: self.__prepare_value(value) for key, value in values.items() if key in keys}


    def __prepare_value(self, value):
        """
            Prevent's saving list of single values.
        """
        if isinstance(value, list) and len(value) == 1:
            return value[0]
        
        return value


    # -------------
    # Setter/-Getter
    # ------------------

    def get_path(self):
        return self.BASE_PATH



def save_history(history, path, filename):
    """
        Saves values of history to the path.
    """
    metrics = Metric(path)
    metrics.write(history, filename)


def read_history(path, filename):
    """
        Reads values from the saved history.
    """
    metrics = Metric(path)
    return metric.read(filename)


def aggregates_per_key(history):
    """
        Aggregate values per key
    """
    
    if len(history) == 0:
        return history

    sample_entry = history[0]
    keys = list(sample_entry.keys())
    aggregates = {key: [] for key in keys}

    for entry in history:
        for key, value in entry.items():
            aggregates[key].append(value)

    return aggregates


import os, sys, select
import numpy as np
import tensorflow as tf

from . import ActiveLearningLoop, AcquisitionFunction
from .wrapper import Model
from .utils import setup_logger

class ExperimentSuit:
    """
    Performs a number of experiments.
    Iterating over given models and methods.

    Parameters:
        models (list(Model)): The models to iterate over.
        query_fns (list(str)|list(AcquisitionFunction)|str|AcquisitionFunction): A list of query functions to use
        dataset (Dataset): A dataset for experiment execution.
        max_rounds (int): The max. number of rounds to query for datapoints per experiment run. If not set, perform query operation as long as there is data.
        acceptance_timeout (int): Timeout in seconds in which experiment can be proceeded or aborted, after successfull (model,query function) iteration. Setting None will automatically proceed. (default: None)
        metrics_handler (ExperimentSuitMetrics): A configured metrics handler to use.
        verbose (bool): Printing log messages?
    """

    def __init__(
        self, 
        models,
        query_fns,
        dataset,
        step_size=1,
        max_rounds=None,
        runs=1,
        seed=None,
        no_save_state=False,
        acceptance_timeout=None,
        metrics_handler=None,
        verbose=False
    ): 

        self.verbose = verbose
        self.logger = setup_logger(verbose, name="ExperimentSuit Logger")

        self.dataset = dataset
        self.max_rounds = max_rounds
        self.runs = runs
        self.step_size = step_size
        self.acceptance_timeout = acceptance_timeout
        self.seed = seed

        self.models = self.__init_models(models)
        self.query_functions = self.__init_query_fns(query_fns)
        self.metrics_handler = metrics_handler
        self.no_save_state = no_save_state
        # os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1' # To disable tensorflow output


    def start(self):
        """
            Starts the experiment suit. 
            Runs an experiment for each acquisition function and model combination.

            TODO:
                [x] Last iteration even when no other experiments to run, prompts proceeding request.
                [ ] Implement run/seed implementation. Run seeds experiments with seeds n-times.
        """
        
        if self.runs == 1:
            self.__iterate_seeds(self.runs-1)
            return


        for run in range(self.runs):
            self.__iterate_experiments(run)


    def __iterate_seeds(self, run):
        """
            Iterate through different seeds when a list of seeds where given.
        """

        if self.seed is not None:

            if isinstance(self.seed, int):
                self.__iterate_experiments(run, self.seed)
                return

            if not isinstance(self.seed, list):
                raise ValueError("Error initializing the active learning loop. Parameter seed of unknown type. Expected list of integers or single integer.")
            
            for idx in range(len(self.seed)):
                seed = self.seed[idx]
                self.__iterate_experiments(idx, seed)

            return
        
        # Run single experiment without seed
        self.__iterate_experiments(run)
    

    def __iterate_experiments(self, run, seed=None):
        """
            
            Parameters:
                run (int): The number of run for acquisition function model and model.

        """

        # Perform experiment for each model & query function combination
        exit_loop = False
        number_of_models = range(len(self.models))
        number_of_query_fns = range(len(self.query_functions))
        for i in number_of_models:
            model = self.models[i]
            
            if not self.no_save_state and not model.has_save_state():
                model.save_weights()

            # Group experiment output per model in terminal
            if i != 0:
                print("#"*10)

            metrics = None
            for j in number_of_query_fns:
                query_fn = self.query_functions[j]

                print("Running experiment (Run: {} | Model: {} | Query-Function: {})".format(run, model, query_fn))
                self.__update_query_function(model, query_fn)
                self.__run_experiment(run, model, query_fn, seed)

                if (j != (len(self.query_functions)-1) or i != (len(self.models)-1)) \
                and not self.__await_proceed():

                    exit_loop = True
                    break
            
            if exit_loop:
                break

    def __run_experiment(self, run, model, query_fn, seed):
        """
            Run a single experiment.

            Parameters:
                run (int): The number of experiment of this type (combination of acquisition funciton and model)
                model (Model): A model wrapper.
                query_ fn (str|AcquisitionFunction): The acquisition function to use.
        """

        # Quick fix, reset random state after each iteration
        # TODO: Adding outter loop and extend parameter list for seed
        if seed is not None and isinstance(seed, int):
            np.random.seed(seed)
            tf.random.set_seed(seed)

        active_learning_loop = ActiveLearningLoop(
            model, 
            self.dataset, 
            query_fn, 
            step_size=self.step_size,
            max_rounds=self.max_rounds,
            pseudo=True,
            verbose=self.verbose
        )

        experiment_name = str(run) + "_" + active_learning_loop.get_experiment_name()
        active_learning_loop.run(experiment_name, self.metrics_handler)


    def __await_proceed(self):
        """
            Waiting for user input to proceed or abort experiments.

            TOOD:
                [ ] Restart user input when failed input
        """

        if self.acceptance_timeout is not None and isinstance(self.acceptance_timeout, int):
            print("Proceed with next experiment? (y/n) ")
            while True:
                i, o, e = select.select([sys.stdin], [], [], 2)

                if i: 
                    value = sys.stdin.readline().strip().lower()
                    if value == "y" or value == "yes":
                        return True
                    elif value == "n" or value == "no":
                        return False
                    else:
                        print("Unknown value passed. Either input yes or no.")
                        continue

                else:
                    print("\033[F Time-out. Auto-proceed with experiments.")
                    return True
            
        return True



    # ------------
    # Utilities
    # ---------------

    def __init_models(self, models):
        """
            Iterate through passed models,
            raising an error when one of the models can't be processed.
        """

        if isinstance(models, Model):
            return [models]

        verified_models = []
        if isinstance(models, list):
            for model in models:
                
                # Passed model can be used in context of ActiveLearningLoop?
                if not isinstance(model, Model):
                    raise ValueError("Error in ExperimentSuit.__init__(). One of the passed models is no sub-class of Model.")

        else:
            raise ValueError("Error in ExperimentSuit.__init__(). Can't parse models of type {}. Expected list of or single Model.".format(type(models)))

        return models


    def __init_query_fns(self, query_fns):
        """
            Create AcquisitionFunction
        """

        if isinstance(query_fns, str) or isinstance(query_fns, AcquisitionFunction):
            return [query_fns]

        fns = []
        if isinstance(query_fns, list):
            for query_fn in query_fns:

                if isinstance(query_fn, str):
                    fns.append(AcquisitionFunction(query_fn))

                elif isinstance(query_fn, AcquisitionFunction):
                    fns.append(query_fn)
                
                else:
                    raise ValueError("Error in ExperimentSuit.__init__(). Can't initialize one of the given AcquisitionFunctions")
        
        else:
            raise ValueError("Error in ExperimentSuit.__init__(). Got type {} for qury_fns. Expected a list of strings, AcqusitionFunctions, singel strings or a single AcquisitionFunction.".format(type(query_fns)))

        return fns


    def __update_query_function(self, model, functions):
        """
            Update the acquisition function to use new model.

            Parameters:
                functions (AcquisitionFunction|list(AcquisitionFunction)): The acquisition functions of the experiment
                model (Model): The model wrapper to use.
        """

        is_str = isinstance(functions, str)
        is_obj = isinstance(functions, AcquisitionFunction)
        is_list = isinstance(functions, list)

        if not (is_obj or is_list or is_str):
            raise ValueError("Error in ActiveLearningLoop.run(). Failed to update acquisition function for model \"{}\".".format(model.__class__.__name__))

        # Model function will be set in active learning loop
        if is_str:
            return

        if is_obj:
            functions._set_fn(model)
            return
        
        # Update a list of AcquisitionFunction objects
        for function in functions:
            function._set_fn(model)


import math
from copy import deepcopy
import numpy as np


class Pool:
    """
        Pool that holds information about labeled and unlabeld inputs.
        The attribute 'indices' holds information about the labeled inputs.
        
        Each value of self.indices can take the following states:
        (value==-1) Corresponding input is labeld
        (value!=-1) Corresponding input is not labeled 

        Parameters:
            inputs (numpy.ndarray): Inputs to the network.
    """

    def __init__(self, inputs, targets=None):
        self.__inputs = inputs
        self.__true_targets = targets
        self.__indices = np.linspace(0, len(inputs)-1, len(inputs), dtype=int)
        self.__targets = np.zeros(len(inputs))


    def init(self, size):
        """
            Initialize the pool with specific number of labels.
            Only applicable when pool in pseudo mode.

            Parameters:
                size (int|list|np.ndarray): Either the number of datapoints to initialized or an explicit list or array of indices to initialize.
        """
        
        is_int = isinstance(size, int)
        is_list = isinstance(size, list)
        is_np_array = isinstance(size, np.ndarray)

        if not self.is_pseudo():
            raise ValueError("Error in Pool.init(size). Can't initialize pool using init(size) when not in pseudo mode. Initialize pool with targets, to put Pool in pseudo mode.")

        if is_int and size < 1:
            raise ValueError("Error in Pool.init(size). Can't initialize pool with {} targets. Use a positive integer > 1.".format(size))

        if is_int and len(self.__indices) < size:
            raise ValueError("Error in Pool.init(size). Can't initialize pool, not enough targets. {} targets required, {} are available.".format(size, len(self.__indices)))

        if not (is_int or is_list or is_np_array):
            raise ValueError("Error in Pool.init(size). Expected size to be an integer, list or numpy array of indices.")

        # Initialize explicit indices
        if is_list or is_np_array:
            self.__init_explicit_indices(size)
            return

        # WARNING: Will only work for categorical targets
        unique_targets = np.unique(self.__true_targets)

        # Initialize n-datapoints per class
        num_to_select = 1
        num_unique_targets = len(unique_targets)
        if num_unique_targets < size:
            num_to_select = math.floor(size/num_unique_targets)    
        
        # Annotate samples in round robin like schme
        while size > 0:
            
            # Select 
            for target in unique_targets:

                unlabeled_indices = self.get_unlabeled_indices()
                true_targets = self.__true_targets[unlabeled_indices]
                selector = (true_targets == target)

                indices = unlabeled_indices[selector]
                targets = true_targets[selector]

                # Move to next target, when none available of this type
                if len(indices) == 0:
                    continue
                
                adapted_num_to_select = self.__adapt_num_to_select(targets, num_to_select)
                selected_indices = np.random.choice(indices, adapted_num_to_select, replace=False)
                
                # Update pool
                selected_targets = self.__true_targets[selected_indices]
                self.annotate(selected_indices, selected_targets)
                size -= num_to_select

                if size < 1:
                    break
    

    def __init_explicit_indices(self, indices):
        """
            Initializes the pool with ex

            Parameters:
                indices (list|numpy.ndarray): A list of indices which to use 
        """        
    
        try:
            unlabeled_indices = self.get_unlabeled_indices()
            true_targets = self.__true_targets[unlabeled_indices]

            selected_indices = unlabeled_indices[indices]
            selected_targets = true_targets[indices]
            self.annotate(selected_indices, selected_targets)

        except IndexError as e:
            raise IndexError("Error in Pool.init(size). " + str(e).capitalize() + ".")

    
    def __adapt_num_to_select(self, available, num_to_select):
        """
            Adapts the number of elements to select next.

            Parameters:
                available (numpy.ndarray): The available elements.
                num_to_select (int): The number of elements to select.

            Returns:
                (int) the adapted number of elements selectable.
        """

        num_available = len(available)
        if num_available < num_to_select:
            return num_available

        return num_to_select


    def get_inputs_by(self, indices):
        """
            Get inputs by indices.

            Parameters:
                indices (numpy.ndarray): The indices at which to access the data.

            Returns:
                (numpy.ndarray) the data at given indices.
        """
        return self.__inputs[indices]


    def get_targets_by(self, indices):
        """

        """
        return self.__targets[indices]


    def __setitem__(self, indices, targets):
        """
            Shortcut to annotate function.

            Parameters:
                indices (numpy.ndarray): For which indices to set values.
                targets (numpy.ndarray): The targets to set.
        """
        self.annotate(indices, targets)

    
    def annotate(self, indices, targets=None):
        """
            Annotate inputs of given indices with given targets.

            Parameters:
                indices (numpy.ndarray): The indices to annotate.
                targets (numpy.ndarray): The labels to set for the given annotations.
        """

        if targets is None:
            if self.__targets is None:
                raise ValueError("Error in Pool.annotate(). Can't annotate inputs, targets is None.")

            targets = self.__true_targets[indices]
            

        # Create annotation
        self.__indices[indices] = -1
        self.__targets[indices] = targets


    # ---------
    # Utilities
    # -------------------

    def has_unlabeled(self):
        """
            Has pool any unlabeled inputs?

            Returns:
                (bool) true or false depending whether unlabeled data exists.
        """
        selector = np.logical_not(self.__indices == -1)
        return np.any(selector)


    def has_labeled(self):
        """
            Has pool labeled inputs?

            Returns:
                (bool) true or false depending whether or not there are labeled inputs.
        """
        selector = self.__indices == -1
        return np.any(selector)


    def is_pseudo(self):
        """
            Is the pool in pseudo mode?
            Meaning, true target labels are already known?

            Returns:
                (bool) indicating whether or not true labels are existent.
        """
        return self.__true_targets is not None


    def __deepcopy__(self, memo):
        return Pool(self.__inputs, self.__true_targets)

    # ---------
    # Setter/Getter
    # -------------------

    def get_indices(self):
        """
            Returns the current labeling state.

            Returns:
                (numpy.ndarray) the indices state. (-1) indicating a labeled input.
        """
        
        return self.__indices


    def get_labeled_indices(self):
        """
            Get the indices of labeled datapoints.
            
            Returns:
                (numpy.ndarray) of datapoints that already has been labeled.
        """
        selector = self.__indices == -1
        indices = np.linspace(0, len(self.__inputs)-1, len(self.__inputs), dtype=int)
        return indices[selector]


    def get_unlabeled_indices(self):
        """
            Get all unlabeled indices for this pool.

            Returns:
                (numpy.ndarray) an array of indices.
        """
        selector = self.__indices != -1
        return self.__indices[selector]


    def get_length_labeled(self):
        """
            Get the number of labeled inputs.

            Returns:
                (int) The number of labeled inputs.
        """
        return np.sum(self.__indices == -1)

    
    def get_length_unlabeled(self):
        """
            Get the number of unlabeld inputs.

            Returns:
                (int) the number of unlabeled inputs
        """
        return np.sum(np.logical_not(self.__indices == -1))   


    def get_labeled_data(self):
        """
            Get data and indices of datapoints which are currently labeled.

            Returns:
                (tuple(numpy.ndarray, numpy.ndarray)) inputs and corresponding targets.
        """
        selector = self.__indices == -1
        inputs = self.__inputs[selector]
        targets = self.__targets[selector]
        return inputs, targets

    
    def get_unlabeled_data(self):
        """
            Get data and their indices of datapoints which are currently not labeled. 

            Returns:
                (tuple(numpy.ndarray, numpy.ndarray)) The inputs and their indices in the pool
        """
        selector = self.__indices != -1
        inputs = self.__inputs[selector]
        indices = self.__indices[selector]
        return inputs, indices


"""
TODO:
        - [ ] Add posibility to add default configurations to be filled if nothing supplied
"""



class Config:
    """
        Configuration to clean up function calls and encapsulate
        connected configurations.

        Args:
            defaults (object): Reserved key for default configuration.
            
    """

    def __init__(self, **kwargs):
        self.kwargs = self.__init_defaults(kwargs)


    def __init_defaults(self, kwargs):
        """
            Set default values for a configuration object.

            Parmeters:
                kwargs (dict): Keyword arguments passed to Config.

            Returns:
                (dict) updated keyword arguments.
        """

        # defaults existing?
        defaults = dict.get(kwargs, "defaults")
        if defaults is None:
            return kwargs

        # Are defaults given as dict?
        if not isinstance(defaults, dict):
            raise ValueError("Can't set defaults. 'defaults' needs to be of type 'dict'.")

        # Set default values
        for key, default_value in defaults.items():
            if dict.get(kwargs, key) is None:
                kwargs[key] = default_value
            
        return kwargs


    # -------------
    # Dunder
    # -------------------

    # Access function
    def __getitem__(self, key):
        return self.kwargs[key]


    def __contains__(self, key):
        return key in self.kwargs


import os, math
import time
import numpy as np
import logging as log
from sklearn.metrics import accuracy_score
import tensorflow.keras as keras

from . import  Model, ModelType, Mode
import tensorflow as tf



class McDropout(Model):
    """
        Wrapper class for neural networks.

    """

    def __init__(self, model, config=None, **kwargs):
        super().__init__(model, config=config, model_type=ModelType.MC_DROPOUT, **kwargs)

        # disable batch norm
        # super().disable_batch_norm()


    def __call__(self, inputs, sample_size=10, batch_size=None, callback=None, **kwargs):
        """
            
            Parameters:
                inputs (numpy.ndarray): Inputs going into the model
                sample_size (int): How many times to sample from posterior?
                batch_size (int): In how many batches to split the data?
        """

        if batch_size is None:
            batch_size = len(inputs)
        
        if batch_size < 1:
            raise ValueError("Error in McDropout.__call__(). Can't select negative amount of batches.")

        if sample_size < 1:
            raise ValueError("Error in McDropout.__call__(). Can't sample negative amount.")


        total_len = len(inputs)
        num_batches = math.ceil(total_len/batch_size)
        batches = np.array_split(inputs, num_batches, axis=0)
        predictions = []

        for batch in batches:
            # Sample from posterior
            posterior_samples = []
            for i in range(sample_size):
                posterior_samples.append(self._model(batch, training=True))
                
            # Omit sample dimension, when only sampled single time?
            if sample_size > 1:
                stacked = np.stack(posterior_samples, axis=1)
                predictions.append(stacked)
            else:
                predictions.append(posterior_samples[0])


        if len(predictions) == 1:
            return predictions[0]

        return np.vstack(predictions)


    def evaluate(self, inputs, targets, sample_size=10, **kwargs):
        """
            Evaluate a model on given input data and targets.
        """
        
        if len(inputs) != len(targets):
            raise ValueError("Error in McDropout.evaluate(). Targets and inputs not of equal length.")

        # Returns: (batch_size, sample_size, target_len) or (batch_size, target_len)
        predictions = self.__call__(inputs, sample_size=sample_size, **kwargs)
        loss, acc = self.__evaluate(predictions, targets, sample_size)
        return {"loss": loss, "accuracy": acc}


    def __evaluate(self, predictions, targets, sample_size):
        """

            Parameters:
                predictions (numpy.ndarray): The predictions made by the network of shape (batch, targets) or (batch, samples, targets)
                targets (numpy.ndarray): The target values
                sample_size (int): The number of samples taken from posterior.

            Returns:
                (list()) of values representing the accuracy and loss
        """
        
        expectation = predictions
        if len(predictions.shape) == 3:
            expectation = np.average(predictions, axis=1)

        # Will fail in regression case!!!! Add flag to function?
        loss_fn = tf.keras.losses.get(self._model.loss)
        loss = loss_fn(targets, expectation)

        # Extend dimension in binary case 
        extended = self.extend_binary_predictions(predictions)
        pred_targets = np.argmax(extended, axis=-1)

        # One-hot vector passed
        if len(targets.shape) == 2:
            targets = np.argmax(targets, axis=1)
        
        # Extend target dimension (multiple sample in prediction)
        if sample_size > 1:
            targets = np.vstack([targets]*sample_size).T
        
        acc = np.mean(pred_targets == targets)
        return [np.mean(loss.numpy()), acc]


    def expectation(self, predictions):
        """
            Calculate the mean of the distribution
            output distribution.

            Returns:
                (numpy.ndarray) The expectation per datapoint
        """
        # predictions -> (batch_size, num_predictions)
        predictions = self.extend_binary_predictions(predictions)
        return np.average(predictions, axis=1)


    def variance(self, predictions):
        """
            Calculate the variance of the distribution.

            Returns:
                (numpy.ndarray) The variance per datapoint and target
        """
        predictions = self.extend_binary_predictions(predictions)
        return np.var(predictions, axis=1)


    def std(self, predictions):
        """
            Calculate the standard deviation.

            Returns:
                (numpy.ndarray) The standard deviation per datapoint and target
        """
        predictions = self.extend_binary_predictions(predictions)
        return np.std(predictions, axis=1)


    def just_return(self, predictions):
        return predictions


    def extend_binary_predictions(self, predictions, num_classes=2):
        """
            In MC Dropout case always predictions of shape
            (batch_size, sample_size, classes) for classification 
            or (batch_size, sample_size) for binary/regression case
        """

        # Don't modify predictions shape in regression case
        if not self.is_classification():
            return predictions


        # Binary case: calculate complementary prediction and concatenate
        if self.is_binary():
            bin_alt_class = (1 + np.zeros(predictions.shape)) - predictions

            # Expand dimensions for predictions to concatenate. Is this needed?
            bin_alt_class = np.expand_dims(bin_alt_class, axis=-1)
            predictions = np.expand_dims(predictions, axis=-1)

            # Concatenate predictions
            predictions = np.concatenate([predictions, bin_alt_class], axis=len(predictions.shape)-1)
        
        return predictions



    # -----
    # Acquisition functions
    # ---------------------------

    def get_query_fn(self, name):

        if name == "max_entropy":
            return self.__max_entropy
        
        if name == "bald":
            return self.__bald
        
        if name == "max_var_ratio":
            return self.__max_var_ratio

        if name == "std_mean":
            return self.__std_mean

        if name == "margin_sampling":
            return self.__margin_sampling

        return None


    def __max_entropy(self, data, sample_size=10, **kwargs):
        """
            Select datapoints by using max entropy.

            Parameters:
                model (tf.Model) The tensorflow model to use for selection of datapoints
                unlabeled_pool (Pool) The pool of unlabeled data to select
        """        
        # Create predictions
        predictions = self.__call__(data, sample_size=sample_size)
        expectation = self.expectation(predictions)
        
        # Absolute value to prevent nan values and + 0.001 to prevent infinity values
        log_post = np.log(np.abs(expectation) + .001)

        # Calculate max-entropy
        return -np.sum(expectation*log_post, axis=1)


    def __bald(self, data, sample_size=10, **kwargs):
        # TODO: dimensions do not line up in mutli class
        # predictions shape (batch, num_predictions, num_classes)
        predictions = self.__call__(data, sample_size=sample_size)
        posterior = self.expectation(predictions)

        first_term = -np.sum(posterior*np.log(np.abs(posterior) + .001), axis=1)

        # Missing dimension in binary case?
        predictions = self.extend_binary_predictions(predictions)
        
        inner_sum = np.sum(predictions*np.log(np.abs(predictions) + .001), axis=1)
        second_term = np.sum(inner_sum, axis=1)/predictions.shape[1]
        return first_term + second_term


    def __max_var_ratio(self, data, sample_size=10, **kwargs):
        """
            Select datapoints by maximising variation ratios.

            # (batch, predictions, classes) reduce to (batch, predictions (max-class))
            # 1 - (count of most common class / num predictions)
        """
        predictions = self.__call__(data, sample_size=sample_size)
        posterior = self.expectation(predictions)

        # Calcualte max variation rations
        return 1 - posterior.max(axis=1)


    def __std_mean(self, data, sample_size=10,  **kwargs):
        """
           Maximise mean standard deviation.
           Check std mean calculation. Depending the model type calculation of p(y=c|x, w) can differ.
           (Kampffmeyer et al. 2016; Kendall et al. 2015)

           Todo:
            Implement distinction for different model types.
        """
        # TODO: generalize for n-classes For binary classes
        predictions = self.__call__(data, sample_size=sample_size)

        # Calculate variance/standard deviation from samples
        variance = self.variance(predictions)
        std = np.square(variance)

        # Mean over target variables
        return np.mean(std, axis=-1)


    def __margin_sampling(self, data, sample_size=10, **kwargs):
        """
            Select sample which minimize distance between two most probable labels.
            Margin Sampling (MS).
        """
        predictions = self.__call__(data, sample_size=sample_size)
        expectation = self.expectation(predictions)

        indices = np.argsort(expecation)[:, :-2]

    
    def __least_confidence(self, dtaa, sample_size=10, **kwargs):
        """
            Select sample which minimize distance between two most probable labels.
            Margin Sampling (MS).
        """
        predictions = self.__call__(data, sample_size=sample_size)
        expectation = self.expectation(predictions)

        return np.argmin(expecation, axis=1)





from enum import Enum
from sklearn.model_selection import train_test_split
from . import Pool


class Dataset:
    """
        Splits a dataset into tree parts. Train/Test/validation.
        The train split is used for selection of 
        

        Parameters:
            inputs (numpy.ndarray): The model inputs.
            targets (numpy.ndarray): The targets, labels or values.
            init_size (int): The initial size of labeled inputs in the pool.
            train_size (float|int): Size of the train split.
            test_size (float|int): Size of the test split.
            val_size (float|int): Size of the validation split.
    """


    def __init__(
        self, 
        inputs,
        targets,
        test=None,
        val=None,
        init_size=0,
        init_indices=None
        # train_size=.75, 
        # test_size=None, 
        # val_size=None
    ):

        self.pseudo = True
        self.init_size = init_size

        self.x_train = inputs
        self.y_train = targets
        
        if test is not None:
            self.x_test, self.y_test = test
        
        if val is not None:
            self.x_test, self.y_test = val
        


        # if len(inputs) != len(targets):
        #     raise ValueError("Error in Dataset.__init__(). Can't initialize dataset. Length of inputs and targets are not equal.")

        # train_size, test_size, val_size = self.__init_sizes(len(inputs), train_size, test_size, val_size)

        # if train_size == 1 and test_size == 0 and val_size == 0:
        #     self.x_train = inputs
        #     self.y_train = targets

        # else:
        #     self.x_train, x_test, self.y_train, y_test = train_test_split(inputs, targets, train_size=train_size)

        #     if test_size != 0 and test_size + train_size == 1:
        #         self.x_test = x_test
        #         self.y_test = y_test

        #     else:
        #         pass


    def __init_sizes(self, data_size, train_size, test_size, val_size):
        """
            Initialize sizes for the different sets.
            Throwing errors when catching sight of disallowed situations.

            Parameters:
                data_size (int): the size of the dataset.
                train_size (int): the size of the training set.
                test_size (int): the size of the test set.
                val_size (int): the size of the evaluate set.

            Return:
                (tuple) of sizes (train_size, test_size, val_size)
        """

        # 1. Cast everything to procentual proportion of dataset
        train_size = self.__cast_to_float(data_size, train_size)
        test_size = self.__cast_to_float(data_size, test_size, "test")
        val_size = self.__cast_to_float(data_size, val_size, "evaluate")

        # [(train, .75), (test, .25), (eval, .0)]
        sets = []
        if train_size is not None:
            sets.append(("train", train_size))
        
        if test_size is not None:
            sets.append(("test", test_size))

        if val_size is not None:
            sets.append(("eval", val_size))

        # No splits of data?
        num_of_sets = len(sets)
        if num_of_sets == 0:
            raise ValueError("Error in Dataset.__init__(). No ")

        # Sum percentages
        # for set_idx in range(num_of_sets):
        #     set_name, set_size = 


    def __cast_to_float(self, total_size, part_size, set_name="train"):
        """
            Transform integer set size into float set size.

            Parameters:
                total_size (int): The dataset size.
                part_size (int|float): The amount of datapoints to select from the dataset.
                set_name (str): The set for which to transform the set size.
            
            Returns:
                (float) the transformed set size, as percentual part of the dataset.
        """

        is_float = isinstance(part_size, float)
        is_integer = isinstance(part_size, int)
        if (is_float or is_integer) and part_size < 0:
            raise valueError("Error in Dataset.__init__(). Can't select negative number of datapoints for {} set.".format(set_name))

        percentage = 0
        if is_float:
            percentage = part_size
        
        elif is_integer:
            percentage = part_size/total_size
            return part_size/total_size
            
        else:
            return .0

        # More than 100% to select?
        if percentage > 1.0:
                raise ValueError("Error in Dataset.__init__(). Can't select more than datapoints than available for {} set.".format(set_name))
        
        return percentage

    # ----------
    # Utilities
    # -------------------

    def is_pseudo(self):
        return self.pseudo

    def has_test_set(self):
        return hasattr(self, 'x_test') and hasattr(self, 'y_test')

    def has_eval_set(self):
        return hasattr(self, 'x_val') and hasattr(self, 'y_val')


    def get_split_ratio(self):
        """

            Returns:
                (int, int, int) the split ratio between (train, test, eval) sets.  
        """

        train_size = len(self.x_train)
        test_size = 0
        if hasattr(self, "x_test"):
            test_size = len(self.x_test)
        
        eval_size = 0
        if hasattr(self, "x_eval"):
            eval_size = len(self.x_eval)
        
        return (train_size, test_size, eval_size)


    def percentage_of(self, total_number, part):
        """
            Calculates the percentage a part takes from given total number.

            Parameters:
                total_number (int): The total number from which to calculate the percentual part.
                part (int): The part of which to calculate the percentage.

            Returns:
                (float) representing the percentage of given part im total number.
        """
        return part/total_number

    
    def check_int_in_range(self, value):
        """

        """
        pass


    def check_float_range(self, value):
        """
            Is float in procentual range?

            Parameters:
                value (float): The value to perform the check on.
        """
        return (value < 1) and (value > 0 )


    # -------------
    # Setter/-Getter
    # -------------------

    def get_init_size(self):
        return self.init_size

    def get_train_split(self):
        return (self.x_train, self.y_train)

    def get_train_inputs(self):
        return self.x_train

    def get_train_targets(self):
        return self.y_train

    def get_test_split(self):
        return (self.x_test, self.y_test)

    def get_eval_split(self):
        return (self.x_val, self.y_val)




import os, time, gc
from copy import copy, deepcopy
from tqdm import tqdm

from .utils import setup_logger, log_mem_usage
from . import AcquisitionFunction, Pool, Oracle, ExperimentSuitMetrics


class ActiveLearningLoop:
    """
        Creates an active learning loop. The loop accumulates metrics during training in a dictionary
        that is returned.

        To use with tqdm: 
        ::
            for i in tqdm(my_iterable):
            do_something()

        Parameters:
            model (Model): A model wrapped into a Model type object.
            dataset (Dataset): The dataset to use (inputs, targets)
            query_fn (list(str)|str): The query function to use.

        
        **kwargs:
            step_size (int): How many new datapoints to add per iteration.
            init_size (int): The initial labeled pool size.


        Returns:
            (dict()) containing accumulated metrics.
    """

    def __init__(
        self,
        model,
        dataset,
        query_fn,
        step_size=1,
        max_rounds=None,
        pseudo=True,
        verbose=False,
        **kwargs
    ):
        
        self.verbose = verbose
        self.logger = self.logger = setup_logger(verbose, "ActiveLearningLoop-Logger")

        # Data and pools
        self.dataset = dataset
        x_train, y_train = dataset.get_train_split()
        self.initial_size = initial_pool_size = dataset.get_init_size()

        self.pool = Pool(x_train, y_train)
        if dataset.is_pseudo() and initial_pool_size > 0:
            self.pool.init(initial_pool_size)
        
        # Loop parameters
        self.step_size = step_size
        self.iteration_user_limit = max_rounds
        self.iteration_max = int(self.pool.get_length_unlabeled())
        self.i = 0

        # Active learning components
        self.model = model
        self.oracle = Oracle(pseudo_mode=pseudo)
        self.query_fn = self.__init_acquisition_fn(query_fn)

        self.query_config = self.model.get_query_config()
        self.query_config.update({"step_size": step_size})


    def __len__(self):
        """
            How many iterations until the active learning loop exits.

            Returns:
                (int) The number of iterations.
        """

        times, rest = divmod(self.iteration_max, self.step_size)
        if rest != 0:
            times += 1

        if self.iteration_user_limit is not None and self.iteration_user_limit < times:
            return self.iteration_user_limit
        
        return times


    # ---------
    # Iterator Protocol
    # -----------------

    def __iter__(self):
        self.i = 0
        return self


    def __next__(self):
        """
            Iterate over dataset and query for labels.

            Returns:
                (dict) Accumulated information during avtive learning iterations.
        """

        if not self.has_next():
            raise StopIteration

        # Load previous checkpoints/recreate model
        self.logger.info("++++++++++++++ (START) Iteration {} ++++++++++++++".format(self.i))
        self.model.reset(self.pool, self.dataset)
        
        # Optimiize model params
        optim_metrics, optim_time = self.__optim_model_params()

        # Fit model
        train_metrics, train_time = self.__fit_model()
        self.logger.info("\\---- (END) Model fitting")

        # Update pools
        acq_start = time.time()
        self.logger.info("\\---- (START) Acquisition")
        indices, _pred = self.query_fn(self.model, self.pool, **self.query_config)
        self.logger.info("\\---- (END) Acquisition")
        acq_time = time.time() - acq_start

        self.oracle.annotate(self.pool, indices)

        # Evaluate model
        eval_metrics, eval_time = self.__eval_model()
        self.i += 1

        self.logger.info("/// Memory ")
        gc.collect()
        self.model.clear_session()
        log_mem_usage(self.logger)
        self.logger.info("++++++++++++++ (END) Iteration ++++++++++++++")

        return {
            "train": train_metrics,
            "train_time": train_time,
            "query_time": acq_time,
            "optim": optim_metrics,
            "optim_time": optim_time,
            "eval": eval_metrics,
            "eval_time": eval_time,
            "indices_selected": indices.tolist()
        }


    def __optim_model_params(self):
        """
            Perform parameter optimization using on a validation set.
        """

        metrics = None
        duration = None
        if hasattr(self.model, "optimize") and self.dataset.has_eval_set():
            e_inputs, e_targets = self.dataset.get_eval_split()
            start = time.time()
            metrics = self.model.optimize(e_inputs, e_target)
            duration = time.time() - start

        return metrics, duration

    
    def __fit_model(self):
        """
            Fit model to the labeled data.

            Returns:
                (tuple(dict(), float)) metrics and the time needed to fit the model.
        """
        history = None
        duration = None

        self.logger.info("//// (START) Model fitting")
        if self.pool.get_length_labeled() > 0:
            inputs, targets = self.pool.get_labeled_data()
            start = time.time()

            h = None
            if self.dataset.has_eval_set():
                x_eval, y_eval = self.dataset.get_eval_split()
                h = self.model.fit(inputs, targets, validation_data=(x_eval, y_eval), verbose=False)
            else:
                h = self.model.fit(inputs, targets, verbose=False)

            duration = time.time() - start
            history = h.history

        return history, duration


    def __eval_model(self):
        """
            Performan an evaluation of the model.

            Returns:
                (tuple(dict(), float)) metrics and the time needed to evaluate the model.
        """
        metrics = None
        duration = None
        config = self.model.get_eval_config()

        # print("Config: {}".format(self.model.get_config()))
        if self.dataset.has_test_set():
            x_test, y_test = self.dataset.get_test_split()
            start = time.time()
            metrics = self.model.evaluate(x_test, y_test, **config)
            duration = time.time() - start
        
        return metrics, duration


    # -------
    # Functions for diverse grades of control
    # ---------------------------------

    def run(self, experiment_name=None, metrics_handler=None):
        """
            Runs the active learning loop till the end.

            Parameters:
                experiment_name (str): The name of the file to write to
                metrics_handler (ExperimentSuitMetrics): Metrics handler for write/read operations.
        """

        if experiment_name is None:
            experiment_name = self.get_experiment_name()

        # Write meta information
        if metrics_handler is not None \
        and isinstance(metrics_handler, ExperimentSuitMetrics):
            model_name = self.get_model_name()
            query_fn = self.get_query_fn_name()
            params = self.collect_meta_params()
            metrics_handler.add_experiment_meta(experiment_name, model_name, query_fn, params)

        iteration = 0
        with tqdm(total=self.__len__()) as pbar:
            for metrics in self:
                iteration += 1
                pbar.update(1)

                metrics.update({
                    "iteration": iteration,
                    "labeled_pool_size": self.pool.get_length_labeled()-self.step_size,
                    "unlabeled_pool_size": self.pool.get_length_unlabeled()
                })

                # Write metrics to file
                if metrics_handler is not None \
                and isinstance(metrics_handler, ExperimentSuitMetrics):
                    metrics_handler.write_line(experiment_name, metrics)


    def step(self):
        """
            Perform a step of the active learning loop.
        """
        return next(self)

    
    def has_next(self):
        """
            Can another step of the active learning loop be performed?
        """

        # Limit reached?
        if (self.iteration_user_limit is not None) and not (self.i < self.iteration_user_limit):
            return False

        if self.i >= self.iteration_max:
            return False

        # Any unlabeled data left?
        if not self.pool.has_unlabeled():
            return False

        return True


    # ----------
    # Utils
    # ----------------

    def is_done(self):
        """
            The active learning has executed and is done.

            Returns:
                (bool) whether or not the loop has executed.
        """
        return not self.has_next()


    def collect_meta_params(self):
        """
            Collect meta information about experiment to be written into .meta.json.

            Returns:
                (dict) with all meta information.
        """

        iterations = self.iteration_max
        if self.iteration_user_limit is not None and self.iteration_user_limit < self.iteration_max:
            iterations = self.iteration_user_limit

        # fitting_params = model.get_compile_params()
        initial_indices = []
        if self.pool.has_labeled():
            initial_indices = self.pool.get_labeled_indices().tolist()


        return {
            "iterations": iterations,
            "step_size": self.step_size,
            "initial_size": self.initial_size,
            "initial_indices": initial_indices
        }

    
    # -----------
    # Initializers
    # --------------------

    def __init_acquisition_fn(self, functions):

        # Single acquisition function?
        if isinstance(functions, str):
            return AcquisitionFunction(functions)

        # Already acquisition function
        if isinstance(functions, AcquisitionFunction):
            return functions

        else:
            raise ValueError(
                "Error in ActiveLearningLoop.__init_acquisition_fn(). Can't initialize one of given acquisition functions. \
                Expected value of type str or AcquisitionFunction. Received {}".format(type(functions))
            )
    

    # --------
    # Getters-/Setter
    # ------------------
    
    def get_experiment_name(self):
        return self.get_model_name() + "_" + self.get_query_fn_name()

    def get_model_name(self):
        return self.model.get_model_name().lower()
    
    def get_query_fn_name(self):
        return self.query_fn.get_name()



class Stat:
    """
        Use active learning output to compute statistically interesting stuff.

    """

    def __init__(self):
        pass


    @staticmethod
    def jensen_shannon():
        pass

    
    @staticmethod
    def labeling_efficiency(baseline, alternative_method, lable_size):
        """
            Calculate the labeling efficiency for a baseline value (usually random selection) against
            an alternative query method, for a specific labeled pool size.

            Parameters:
                baseline (float|numpy.ndarray): The accuracy of the baseline. Either single value or multiple values for different label sizes.
                alternative_method (float|numpy.ndarray): The accuracy of the alterantive method. Either single value or multiple values.
                label_size (int|numpy.ndarray): The size of the labeled pool

            Returns:
                (float|numpy.ndarry) single value or multiple labeling efficiencies, depending the given inputs.
        """
        pass

    
    @staticmethod
    def auc():
        pass


import os
import tensorflow as tf

def setup_growth():
    """
        Setup memory to grow. Check tf.config.experimental.set_memory_growth for reference.
    """

    gpus = tf.config.experimental.list_physical_devices("GPU")
    if gpus:
        try:
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
                
            logical_gpus = tf.config.experimental.list_logical_devices("GPU")
            print(len(gpus), "Physical GPU's, ", len(logical_gpus), "Logical GPU's")
            
        except RuntimeError as e:
            print(e)
            
    else:
        cpus = tf.config.experimental.list_physical_devices("CPU")
        try:
            logical_cpus = tf.config.experimental.list_logical_devices("CPU")
            print(len(cpus), "Physical CPU,", len(logical_cpus), "Logical CPU")
            
        except RuntimeError as e:
            print(e)


def set_tf_log_level(level="2"):
    """
        Set a log level for tensorflow logging messages.

        Parameters:
            level (str): The log level, one of [0, 1, 2, 3].
    """
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = level


def disable_tf_logs():
    """
        Disable tensorflow log messages.
    """
    set_tf_log_level()




class Experiment:

    """
        A single runable experiment.
    """


    def __init__(self, model, query_fn, dataset=None, **kwargs):
        self.__model = model
        self.__query_fn = query_fn
        self.__dataset = dataset


    def run(self, dataset=None):
        """
            Runs the experiment.

            Parameters:
                dataset (Dataset): The dataset on which to perform active learning.

            
            Returns:
                (Metrics)
        """

        if not self.has_dataset() and dataset is None:
            raise ValueError("Can't perform experiment without a dataset.")

        pass



    # ----------
    # Setter/-Getter
    # -----------------

    def get_dataset(self):
        return self.__dataset

    def get_query_fn(self):
        return self.__query_fn

    def get_model(self):
        return self.__model


    # ---------
    # Uitilities
    # ---------------

    def has_dataset(self):
        return self.__dataset is not None

    





import os, sys, logging


def setup_logger(debug, name="Runner", log_level=logging.DEBUG, default_log_level=logging.CRITICAL):
    """
        Setup a logger for the active learning loop

        Parameters:
            debug (bool): activate logging output in console?
            name (str): The name of the logger to use. (default='Runner')
            log_level (logging.level): The log level to use when debug==True. (default=logging.DEBUG)
            default_log_level (logging.level): The default log level to use when debug==False. (default=logging.CRITICAL)

        Returns:
            (logging.Logger) a configured logger object.
    """

    logger = logging.Logger(name)
    log_level = log_level if debug else default_log_level

    logger.handler = logging.StreamHandler(sys.stdout)
    logger.handler.setLevel(log_level)
    
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    logger.handler.setFormatter(formatter)
    logger.addHandler(logger.handler)
    return logger


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tf_al', 'tf_al.utils', 'tf_al.wrapper']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1,<2',
 'scikit-learn>=0.24.2,<0.25.0',
 'tensorflow>=2,<3',
 'tqdm>=4.62.2,<5.0.0']

setup_kwargs = {
    'name': 'tf-al',
    'version': '0.0.4',
    'description': 'Active learning with tensorflow. Create custom and generic active learning loops. Export and share your experiments.',
    'long_description': '\n\n[![PyPI version](https://badge.fury.io/py/tf-al.svg)](https://badge.fury.io/py/tf-al)\n[![PyPi license](https://badgen.net/pypi/license/tf-al/)](https://pypi.org/project/tf-al/)\n![Python Version: ^3.6](https://img.shields.io/badge/python-%5E3.6-blue)\n\n\n# Active learning with tensorflow\n\n<sup>*Currently only supports classification tasks.</sup>\n\nPerform active learning in tensorflow with extendable parts.  \n\n\n\n\n# Index\n\n1. [Installation](#Installation)\n2. [Documentation](https://exleonem.github.io/tf-al/)\n3. [Getting started](#Getting-started)\n    1. [Model wrapper](#Model-wrapper)\n    2. [Acquisition functions](#Acquisition-functions)\n    1. [Basic active learning loop](#Basic-active-learning-loop)\n4. [Development](#Development)\n    1. [Setup](#Setup)\n    2. [Scripts](#Scripts)\n5. [Contribution](#Contribution)\n6. [Issues](#Issues)\n\n\n\n# Dependencies\n\n```toml\npython="^3.6"\ntensorflow="^2.0.0"\nscikit-learn="^0.24.2"\nnumpy="^1.0.0"\ntqdm="^4.62.6"\n```\n\n# Installation\n\n\n```shell\n$ pip install tf-al\n```\n\n<sup>*To use a specific version of tensorflow or if you want gpu support you should manually install tensorflow. Else this package automatically will install the lastest version of tensorflow described in the [dependencies](#Dependencies).</sup>\n\n# Getting started\n\n\nFollowing the active learning paradigm the most essential parts are the model and the pool of labeled/unlabeled data.\n\n\nTo enable modularity tensorflow models are wrapped. The model wrapper acts as an interface between the active learning loop and the model.\nIn essence the model wrapper defines methods which are called at different steps in the active learning loop.\nTo manage the labeled and unlabeled datapoints the pool class can be used. Which offers methods to label and select datapoints, labels and indices.\n\n\nOther parts provided by the library easy the setup of active learning loops. The active learning loop class uses a dataset and model to creat an iterator, which then can be used to perform active learning over a single experiment.(model and query strategy combination)\n\nThe experiment suit can be used to perform a couple of experiments in a row, which is useful if for example you want to compare differnt acquisition functions.\n\n\n## Model wrapper\n\nModel wrappers are used to create an interface between the tensorflow model and the active learning loop.\nCurrently there are two wrappers defined. `Model` and `McDropout` for bayesian active learning. \nThe `Model` wrapper can be used to create custom model wrappers.\n\n\nHere is an example of how to create and wrap a basic McDropout model.\n\n```python\nfrom tensorflow.keras import Sequential\nfrom tensorflow.keras.layers import Conv2D, MaxPooling2D, Dropout, Dense, Input, Flatten\nfrom tf_al.wrapper import McDropout\n\n# Define and wrap model (here McDropout)\nbase_model = Sequential([\n    Conv2D(32, 3, activation=tf.nn.relu, padding="same", input_shape=input_shape),\n    Conv2D(64, 3, activation=tf.nn.relu, padding="same"),\n    MaxPooling2D(),\n    Dropout(.25),\n    Flatten(),\n    Dense(128, activation=tf.nn.relu),\n    Dropout(.5),\n    Dense(output, activation="softmax")        \n])\n\n# Wrap, configure and compile\nmodel_config = Config(\n    fit={"epochs": 200, "batch_size": 10},\n    query={"sample_size" 25},\n    eval={"batch_size": 900, "sample_size": 25}\n)\nmodel = McDropout(base_model, config=model_config)\nmodel.compile(\n    optimizer="adam", \n    loss="sparse_categorical_crossentropy", \n    metrics=[keras.metrics.SparseCategoricalAccuracy()]\n)\n```\n\n\n### Basic methods\n\n\nThe model wrapper in essence can be used like a regular tensorflow model.\n\n```python\nmodel = McDropout(base_model)\nmodel.compile(optimizer="adam", loss="sparse_categorical_crossentropy", metrics=[keras.metrics.SparseCategoricalAccuracy()])\n\n\n# Fitting the model\nmodel.fit(inputs, targets, batch_size=25, epochs=100)\n\n# Evaluating\nmodel.evaluate(some_inputs, some_targets)\n\n# Predicting\nmodel(inputs, **additional_params)\n```\n\nTo define a custom  custom model wrapper, simply extend your own class using the `Model` class and \noverwrite functions as needed. The regular tensorflow model can be accessed via `self._model`.\n\nTo provide your model wrappers as a package you can simply use the [template on github](https://github.com/ExLeonem/tf-al-ext), which already offers a poetry package setup.\n\n\n\n```python\nfrom tf_al import Model\n\n\nclass CustomModel(Model):\n\n    def __init__(self, model, **kwargs):\n        super().__init__(model, **kwargs)\n\n\n    def __call__(self, *args, **kwargs):\n        # Custom __call__ or standard tensorflow __call__\n\n\n    def predict(self, inputs, **kwargs):\n        # Custom prediction method or the standard tensorflow call model(inputs)\n        \n\n    def evaluate(self, inputs, targets, **kwargs):\n        # Defining custom evaluate method\n        # else standard evaluate method of tensorflow used.\n        return {"metric_1": some_value, "metrics_2": some_other_value}\n\n\n    def fit(self, *args, **kwargs):\n        # Custom fitting procedure, else tensorflow .fit() method is used. \n        \n\n    def compile(self, *args, **kwargs):\n        # Custom compile method else using tensorflow .compile(**kwargs)\n        \n\n    def reset(self, pool, dataset):\n        # In Which way to reset the network after each active learning round\n        # standard is re-loading weights when enabled\n```\n\n\n## Acquisition functions\n\n\n\n## Basic active learning loop\n\n\n```python\n\nimport tensorflow.keras as keras\nfrom tensorflow.keras import Sequential\nfrom tensorflow.keras.layers import Conv2D, MaxPooling2D, Dropout, Dense, Input, Flatten\n\nfrom tf_al import ActiveLearningLoop, Dataset\nfrom tf_al.wrapper import McDropout\n\n# Load dataset and pack into dataset\n(x_train, y_train), test_set = keras.datasets.mnist.load()\ninital_pool_size = 20\ndataset = Dataset(x_train, y_train, test=test_set, init_size=initial_pool_size)\n\n# Create and wrap model\nbase_model = Sequential([\n    Conv2D(32, 3, activation=tf.nn.relu, padding="same", input_shape=input_shape),\n    Conv2D(64, 3, activation=tf.nn.relu, padding="same"),\n    MaxPooling2D(),\n    Dropout(.25),\n    Flatten(),\n    Dense(128, activation=tf.nn.relu),\n    Dropout(.5),\n    Dense(output, activation="softmax")        \n])\n\nmc_model = McDropout(base_model)\nmc_model.compile(\n    optimizer="adam", \n    loss="sparse_categorical_crossentropy", \n    metrics=[keras.metrics.SparseCategoricalAccuracy()]\n)\n\n# Create and start experiment suit (Collection of different experiments model + query_strategy)\nquery_strategy = "random"\nactive_learning_loop = ActiveLearningLoop(\n    mc_model,\n    dataset,\n    query_strategy,\n    step_size=10, # Number of new datapoints to select after each round\n    max_rounds=100 # How many active learning rounds per experiment?\n)\n\n# To completely run through the active learning loop\nactive_learning_loop.run()\n\n# Manually iterate over active learning loop\nfor step in active_learning_loop:\n\n    # Dict with accumulated metrics \n    # ["train", "train_time", "query_time", "optim", "optim_time", "eval", "eval_time", "indices_selected"]\n    step["train"]\n\n\n# Alternativly iterate step inside the loop\nnum_rounds = 10\nfor i in range(num_rounds):\n\n    metrics = active_learning_loop.step()\n    # ... do something with the metrics\n```\n\n## Basic experiment suit setup\n\n```python\nimport tensorflow as tf\nfrom tensorflow.keras import Model, Sequential\nfrom tensorflow.keras.layers import Conv2D, MaxPooling2D, Dropout, Dense, Input, Flatten\n\nfrom tf_al import ActiveLearningLoop, Dataset, Config, ExperimentSuit, AcquisitionFunction\nfrom tf_al.wrapper import McModel\n\n# Split data and put into a dataset\nx_train, x_test, y_train, y_test = train_test_split(some_inputs, some_targets, test_size=test_set_size)\n\n# Number of initial datapoints in pool of labeled data\ninitial_pool_size = 20 \ndataset = Dataset(\n    x_train, y_train,\n    test=(x_test, y_test),\n    init_size=initial_pool_size\n)\n\n# Define and wrap model (here McDropout)\nbase_model = Sequential([\n    Conv2D(32, 3, activation=tf.nn.relu, padding="same", input_shape=input_shape),\n    Conv2D(64, 3, activation=tf.nn.relu, padding="same"),\n    MaxPooling2D(),\n    Dropout(.25),\n    Flatten(),\n    Dense(128, activation=tf.nn.relu),\n    Dropout(.5),\n    Dense(output, activation="softmax")        \n])\n\nmodel_config = Config(\n    fit={"epochs": 200, "batch_size": 10}, # Passed to fit() of the wrapper\n    query={"sample_size" 25}, # Configuration passed to acquisition function during query step\n    eval={"batch_size": 900, "sample_size": 25} # Parameters passed to evaluation method of the wrapper\n)\nmodel = McDropout(base_model, config=model_config)\nmodel.compile(\n    optimizer="adam", \n    loss="sparse_categorical_crossentropy", \n    metrics=[keras.metrics.SparseCategoricalAccuracy()]\n)\n\n# Over which model to perform experiments single or list [model_1, ..., model_n]\nmodels = model\n\n# Define which acquisition functions to apply in separate runs either single one or a list [acquisition_1, ...] \nacquisition_functions = ["random", AcqusitionFunction("max_entropy", batch_size=900)]\nexperiments = ExperimentSuit(\n    models,\n    acquisition_functions,\n    step_size=10, # Number of new datapoints to select after each round\n    max_rounds=100 # How many active learning rounds per experiment?\n)\n\n```\n\n# Development\n\n\n## Setup\n\n1. Fork and clone the forked repository\n2. Create a virtual env (optional)\n3. [Install and Setup Poetry](https://python-poetry.org/docs/#installation)\n4. Install package dependencies [using poetry](https://python-poetry.org/docs/cli/#install) or set them up manually\n5. Start development\n\n\n## Scripts\n\n### Create documentation\n\nTo create documentation for the `./tf_al` directory. Execute following command\nin `./docs`\n\n```shell\n$ make html\n```\n\nTo clear the generated documentation use following command.\n\n```shell\n$ make clean\n```\n\n\n### Run tests\n\nTo perform automated unittests run following command in the root package directory.\n\n```shell\n$ pytest\n```\n\nTo generate additional coverage reports run.\n\n```shell\n$ pytest --cov\n```\n\n\n# Contribution\n\n\n# Issues',
    'author': 'Maksim Sandybekov',
    'author_email': 'maksim.sandybekov@live.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ExLeonem/tf-al',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import numpy as np

def init_pools(unlabeled_pool, labeled_pool, targets, num_init_per_target=10):
    """
        Initialize the pool with randomly selected values.

        Paramters:
            unlabeled_pool (UnlabeldPool): Pool that holds information about unlabeld datapoints.
            labeled_pool (LabeledPool): Pool that holds information about labeled datapoints.
            targest (numpy.ndarray): The labels of input values.
            num_init_per_target (int): The initial labels used per target. 

        Todo:
            - Make it work for labels with additional dimensions (e.g. bag-of-words, one-hot vectors)
    """

    # unlabeled_pool.update(indices)
    # labeled_pool[indices] = labels

    # Use initial target values?
    if num_init_per_target <= 0:
        return

    # Select 'num_init_per_target' per unique label 
    unique_targets = np.unique(targets)
    for idx in range(len(unique_targets)):

        # Select indices of labels for unique label[idx]
        with_unique_value = targets == unique_targets[idx]
        indices_of_label = np.argwhere(with_unique_value)

        # Set randomly selected labels
        selected_indices = np.random.choice(indices_of_label.flatten(), num_init_per_target, replace=True)

        # WILL NOT WORK FOR LABELS with more than 1 dimension
        unlabeled_pool.update(selected_indices)
        new_labels = np.full(len(selected_indices), unique_targets[idx])
        labeled_pool[selected_indices] = new_labels


import os
import csv, json
import logging

from .utils import setup_logger


class ExperimentSuitMetrics:
    """
        Uses the given path to write and read experiment metrics
        and meta information.

        If the last segment of the path is not existent it will be created.

        Creating a new object pointing to an already existing metrics path will
        reconstruct all metrics files that were written. 

        WARNING: The reconstructred files will be locked for appending and writing. Can be unlocked by using the unlock() method.

        Parameters:
            base_path (str): Where to save the experiments? No recursive creation of directories.
            verbose (bool): Set debugg mode?
    """

    def __init__(self, base_path, verbose=False):
        self.logger = setup_logger(verbose, name="ExperimentSuitMetrics", default_log_level=logging.WARN)

        self.BASE_PATH = base_path
        self.META_FILE_PATH = os.path.join(base_path, ".meta.json")
        self.__setup_dir(base_path)

        # Keep track of written experiment metrics (Code 0=File was loaded, 1=File created)
        self.experiment_files = {}

        # CSV Parameters
        self.delimiter = " "
        self.quotechar = "\""
        self.quoting = csv.QUOTE_MINIMAL
        self.__load_experiments()


    def __setup_dir(self, path):
        """
            Setup a directory for a suit of experiment metrics.
        """

        # Try to create directory if non existen
        if not os.path.exists(path):
            os.mkdir(path)
        
        # Create non-existent meta.json file
        if not os.path.exists(self.META_FILE_PATH):
            # base_content = {"models": [], "dataset": {}, "params": {}, "acquisition_function": [], "run": []}
            base_content = {"experiments": []}
            self.write_meta(base_content)



    def add_dataset_meta(self, name, path, train_size, test_size=None, val_size=None):
        """
            Adding meta information about the dataset used for the experiments

            Parameters:
                name (str): The name of the dataset.
                path (str): The path to the dataset used.
                train_size (float|int): Similiar to sklearn.model_selection.train_test_split.
                test_size (float|int): the size of the test set.
                val_size (float|int): the size of the validation set.
        """


        meta = self.read_meta()
        meta["dataset"] = {
            "name": name,
            "path": path,
            "train_size": train_size
        }

        if test_size is not None:
            meta["dataset"]["test_size"] = test_size
        

        if val_size is not None:
            meta["dataset"]["val_size"] = val_size

        self.write_meta(meta)


    def add_experiment_meta(self, experiment_name, model_name, query_fn, params):
        """
            Adding meta information about an experiment to the meta file.

            Parameters:
                experiment_name (str): The name of the experiment
                model_name (str): Name of the model used
                query_fn (str): Name of the acquisition function
                params (dict): Dictionary of additional parameters to be saved. Like step_size, iterations, ...
        """
        meta = self.read_meta()
        experiments = meta["experiments"]

        experiments.append({
            "experiment_name": experiment_name,
            "model": model_name,
            "query_fn": query_fn,
            "params": params
        })

        meta["experiments"] = experiments
        self.write_meta(meta)


    # ----------
    # Read/Write files
    # -------------------------

    def write_meta(self, content):
        """
            Writes a dictionary to .meta.json.

            Parameters:
                content (dict): The meta information to be written to .meta.json
        """

        with open(self.META_FILE_PATH, "w") as json_file:
            json_file.write(json.dumps(content, indent=4))


    def read_meta(self):
        """
            Reads the meta information from the .meta.json file.

            Returns:
                (dict) of meta information.
        """
        content = {}
        with open(self.META_FILE_PATH, "r") as json_file:
            content = json_file.read()

        return json.loads(content)

    
    def write_line(self, experiment_name, values, filter_keys=None, filter_nan=True):
        """
            Writes a new line into one of the experiment files. 
            Creating the experiment file if it not already exists.

            Parameter:
                experiment_name (str): The name of the experiment performed.
                values (dict): A dictionary of values to write to the experiment file.
                filter_keys (list(str)): A list of str keys to filter keys of given values dictionary.
        """

        # Filter out empty values
        if filter_nan and isinstance(values, dict):
            values = dict(filter(lambda elem: elem[1] is not None, values.items()))

        values = self._resolve_dict(values)
        
        # Filter specific keys
        if filter_keys is not None and isinstance(filter_keys, str):
            values = {key: values[key] for key in filter_keys}

        filename = self._add_extension(experiment_name, "csv")
        file_path = os.path.join(self.BASE_PATH, filename)

        # Was metrics file reconstructed and is locked?
        if experiment_name in self.experiment_files and self.experiment_files[experiment_name] == 0:
            error_msg = "File {} was reconstructed and is locked. Use .unlock(experiment_name) to open this file up for writing.".format(experiment_name)
            raise ValueError(error_msg)
        
        mode = self.__get_mode(experiment_name)
        with open(file_path, mode) as csv_file:
            fieldnames = list(values.keys())
            csv_writer = self.__get_csv_writer(csv_file, fieldnames)
            
            # Experiment file non-existent? Overwrite mode?
            if (experiment_name not in self.experiment_files) or (self.experiment_files[experiment_name] == 2):
                self.experiment_files[experiment_name] = 1
                csv_writer.writeheader()

            csv_writer.writerow(values)


    def read(self, experiment_name):
        """
            Read metrics from a specific experiment.

            Parameters:
                experiment_name (str): The experiment to read from.

            Returns:
                (list(dict)) of accumulated experiment metrics.
        """

        # .csv extension in filename? 
        experiment_name = self._add_extension(experiment_name, "csv")

        values = []
        experiment_file_path = os.path.join(self.BASE_PATH, experiment_name) 
        with open(experiment_file_path, "r") as csv_file:

            csv_reader = self.__get_csv_reader(csv_file)
            for row in csv_reader:
                values.append(row)
            
        
        return values


    def unlock(self, experiment_name):
        """
            Unlocks a reconstructed file to be available to write it again.

            Parameters:
                experiment_name (str): Name of the expierment to unlock for appending.
        """

        if not experiment_name in self.experiment_files:
            return

        self.experiment_files[experiment_name] = 1


    def unlock_all(self):
        """
            Unlocks all locked files, being able to append to files again.
        """

        for key, value in self.experiment_files.items():
            if value == 0:
                self.experiment_files[key] = 1


    def overwrite(self, experiment_name):
        """
            Mark reconstructed experiment metrics to be overwriten.

            Parameters:
                experiment_name (str): Name of the experiment to mark for overwriting.
        """

        if experiment_name not in self.experiment_files:
            return
        
        self.experiment_files[experiment_name] = 2

        

    # ---------
    # Utilities
    # --------------------
    
    def _resolve_dict(self, values, prefix=None):
        """
            Resolves a dictionary into a flat pandas dataframe like structure.
            Nested dictionaries are getting prefixed with parent key.

            Parameters:
                values (dict): A dictionary of keys. Can include dictionaries with single level nesting.

            Returns:
                (dict) a flattened dictionary.
        """
        
        flattened_dict = {}
        for key, value in values.items():
        
            # Copy flat values into flattened dictionary
            prefixed_key = key if prefix is None else (prefix + "_" + key)
            if not isinstance(values[key], dict):
                flattened_dict[prefixed_key] = value
                continue
            
            resolved = self._resolve_dict(values[key], prefix=prefixed_key)
            flattened_dict.update(resolved)

        return flattened_dict



    def _add_extension(self, filename, ext):
        """
            Adds an extension to a filename.

            Parameters:
                filename (str): The filename to check for the extension
                ext (str): The file extension to add and check for
            
            Returns:
                (str) the file name with a file extension appended. 
        """

        if ext not in filename:
            return filename + "." + ext
        
        return filename


    def __load_experiments(self):
        """
            Reconstrcut metrics from files available files.
        """

        if not os.path.exists(self.BASE_PATH):
            return

        dir_content = os.listdir(self.BASE_PATH)
        for element in dir_content:
            
            # Skip sub-directories
            element_path = os.path.join(self.BASE_PATH, element)
            if not os.path.isfile(element_path):
                continue

            # Skip meta file
            if ".meta.json" in element:
                continue
            
            # Strip extension off of filename
            name, ext = os.path.splitext(element)
            self.experiment_files[name] = 0
        


    # -----------
    # Getter/-Setter
    # ------------------

    def __get_mode(self, experiment_name):
        default_mode = "a"
        if experiment_name not in self.experiment_files:
            return default_mode

        status = self.experiment_files[experiment_name]
        if status == 2:
            return "w"

        return default_mode


    def __get_csv_params(self):
        return {
            "delimiter": self.delimiter,
            "quotechar": self.quotechar,
            "quoting": self.quoting
        }


    def __get_csv_writer(self, file, fieldnames):
        csv_params = self.__get_csv_params()
        return csv.DictWriter(file, fieldnames, **csv_params)

    
    def __get_csv_reader(self, file):
        csv_params = self.__get_csv_params()
        return csv.DictReader(file, **csv_params)


    def get_dataset_info(self):
        """
            Read 

            Returns:
                (dict) containing meta information about the used dataset for the experiment
        """
        
        meta = self.read_meta()
        return meta.get("dataset", None)


    def get_experiment_meta(self, experiment_name):
        """

            Parameter:
                experiment_name (self): The name of the experiment.
        """
        meta = self.read_meta()
        experiments = meta["experiments"]
        experiment_information = {}
        for experiment in experiments:
            
            if experiment["experiment_name"] == experiment_name:
                experiment_information = experiment

        return experiment_information
        


from .checkpoint import Checkpoint
from .utils import disable_batch_norm, predict_n, batch_predict_n, measures
from .model import Model, Mode, ModelType
from .mc_dropout import McDropout
# from .moment_propagation import MomentPropagation


import os, math
import uuid
import numpy as np
from enum import Enum
import tensorflow as tf

from . import Checkpoint
from ..utils import setup_logger


class Mode(Enum):
    TRAIN=1,
    TEST=2,
    EVAL=3


class ModelType(Enum):
    """
        Different bayesian model types.
    """
    MC_DROPOUT=1,
    MOMENT_PROPAGATION=2,
    SWAG=3


class Model:
    """
        Base wrapper for deep learning models to interface
        with the active learning environment. 
        
        Attributes:
            _model (tf.Model): Tensorflow or pytorch module.
            _config (Config): Model configuration
            _mode (Mode): The mode the model is in 'train' or 'test'/'eval'.
            _model_type (ModelType): The model type
            _checkpoints (Checkpoint): Created checkpoints.

        Parameters:
            model (tf.Model): The tensorflow model to be used.
            config (Config): Configuration object for the model. (default=None)
    """

    def __init__(
        self, 
        model, 
        config=None, 
        mode=Mode.TRAIN,
        name=None,
        model_type=None, 
        classification=True, 
        is_binary=False,
        checkpoint=None,
        verbose=False,
        checkpoint_path=None,
        **kwargs
    ):

        self.logger = setup_logger(verbose, "Model Logger")
        self.__verbose = verbose
        self.__id = uuid.uuid1()
        self._model = model
        self._config = config
        self._mode = mode
        self._model_type = model_type
        self._name = name
        

        # Checkpoints path set?
        if checkpoint_path is None:
            checkpoint_path = os.getcwd()
        self._checkpoints = Checkpoint(checkpoint_path) if checkpoint is None else checkpoint

        self.__classification = classification
        if not self.__classification:
            # Binary classification always false, when regression problem
            self.__is_binary = False
        else:
            self.__is_binary = is_binary


    def __call__(self, *args, **kwargs):
        return self._model(inputs, training=self.in_mode(Mode.TRAIN))


    def predict(self, inputs, **kwargs):
        """
            Approximate predictive distribution.

            Parameter:
                inputs (numpy.ndarray): The inputs for the approximation

        """
        return self._model(inputs, training=self.in_mode(Mode.TRAIN))
    

    def evaluate(self, inputs, targets, **kwargs):
        """
            Evaluate a model on given input data and targets.

            Parameters:
                inputs (numpy.ndarray):
                targets (numpy.ndarray):

            Returns:
                (list) A list with two values. [loss, accuracy]  
        """
        loss, accuracy = self._model.evaluate(inputs, targets, **kwargs)
        return {"loss": loss, "accuracy": accuracy}


    def fit(self, *args, **kwargs):
        """
            Fit the model to the given data.

            Args:
                x (numpy.ndarray): The inputs to train the model on. (default=None)
                y (numpy.ndarray): The targets to fit the model to. (default=None)
                batch_size (int): The size of each individual batch

            Returns:
                () a record of the trianing procedure
        """

        if self._config is not None and "fit" in self._config and isinstance(self._config["fit"], dict):
            fit_params = self._config["fit"]
            kwargs.update(fit_params)

        return self._model.fit(*args, **kwargs)


    def compile(self, *args, **kwargs):
        """
            Compile the model if needed
        """
        self._model.compile(**kwargs)


    def prepare_predictions(self, predictions):
        """
            Extend predictions for binary classification case.

            Parameters:
                predictions (numpy.ndarray): The predictions made by the model

            Returns:
                (numpy.ndarray) The extended numpy array
        """
        return predictions

    
    def map_eval_values(self, values):
        """
            Create a dictionary mapping for evaluation metrics.

            Parameters:
                values (any): Values received from model.evaluate

            Returns:
                (dict) The values mapped to a specific key.
        """
        metric_names = self._model.metrics_names
        return dict(zip(metric_names, values))


    def disable_batch_norm(self):
        """
            Disable batch normalization for activation of dropout during prediction.

            Parameters:
                - model (tf.Model) Tensorflow neural network model.
        """
        
        disabled = False
        for l in self._model.layers:
            if l.__class__.__name__ == "BatchNormalization":
                disabled = True
                l.trainable = False

        if disabled:
            self.logger.info("Disabled BatchNorm-Layers.")


    def clear_session(self):
        tf.keras.backend.clear_session()



    # ------
    # Model runtime configurations
    # -----------------------------

    def reset(self, pool, dataset):
        """
            Use to reset states, weights and other stuff during each active learning loop iteration.
        """
        # self.load_weights()
        pass


    def optimize(self, inputs, targets):
        """
        Use to optimize parameters during active learning loop        
        """
        pass


    # ----------------------
    # Utilities
    # ------------------------------

    def batch_prediction(self, inputs, batch_size=1, **kwargs):
        """
            
            Parameters:
                inputs (numpy.ndarray): Inputs going into the model
                n_times (int): How many times to sample from posterior?
                batch_size (int): In how many batches to split the data?
        """
    
        if batch_size < 1:
            raise ValueError("Error in McDropout.__call__(). Can't select negative amount of batches.")

        total_len = len(inputs)
        num_batches = math.ceil(total_len/batch_size)
        batches = np.array_split(inputs, num_batches, axis=0)

        predictions = []
        for batch in batches:
            predictions.append(self._model(batch, training=True))

        return np.vstack(predictions)


    # --------------
    # Checkpoint creation/loading
    # ------------------------------

    def empty_checkpoint(self):
        return self._checkpoints.empty()

    def new_checkpoint(self):
        self._checkpoints.new(self._model)

    
    def load_checkpoint(self, iteration=None):
        self._checkpoints.load(self._model, iteration)


    def clear_checkpoints(self):
        self._checkpoints.clean()


    def save_weights(self):
        path = self._checkpoints.PATH
        self._model.save_weights(path)


    def load_weights(self):
        path = self._checkpoints.PATH
        self._model.load_weights(path)
        

    def has_save_state(self):
        try:
            self.load_weights()
            return True

        except:
            return False



    # -----------
    # Access Configuration
    # ----------------------------

    def get_fit_config(self):
        if (self._config is not None) and "fit" in self._config:
            return self._config["fit"]
        
        return {}


    def get_query_config(self):
        if (self._config is not None) and "query" in self._config:
            return self._config["query"]

        return {}


    def get_eval_config(self):
        if (self._config is not None) and "eval" in self._config:
            return self._config["eval"]
        
        return {}


    def get_config(self):
        return self._config.kwargs



    # --------------
    # Access important flags for predictions
    # -----------------------------

    def in_mode(self, mode):
        return self._mode == mode


    def is_classification(self):
        return self.__classification


    def is_binary(self):
        return self.__is_binary


    # ---------------
    # Acquisition functions
    # --------------------------

    def get_query_fn(self, name):
        """
            Get model specific acquisition function.

            Parameters:
                name (str): The name of the acquisition function to return.

            Returns:
                (function) the acquisition function to use.
        """
        pass


    def __max_entropy(self, data, **kwargs):
        pass

    def __bald(self, data, **kwargs):
        pass

    def __max_var_ratio(self, data, **kwargs):
        pass

    def __std_mean(self, data, **kwargs):
        pass
        

    # -----------------
    # Setter/-Getter
    # --------------------------

    def get_id(self):
        return str(self.__id)

    def get_model_name(self, prefix=True):
        """
            Returns the model name.
        """

        model_type = None
        if not (self._model_type is None) and prefix:
            _pre, model_type = str(self._model_type).split(".")
            model_typ3 = model_type.lower()


        # 
        if not (self._name is None): 
            if not (model_type is None) and prefix:
                return model_type + "_" + self._name

            return self._name

        # No specific name set
        if not (model_type is None):
            return model_type

        return "model"
        

    def get_model_type(self):
        return self._model_type

    def get_model(self):
        return self._model

    def get_mode(self):
        return self._mode

    def get_metric_names(self):
        return self._model.metrics_names

    def set_mode(self, mode):
        self._mode = mode

    def get_base_model(self):
        return self._model

    # ---------------
    # Dunder
    # ----------------------

    def __eq__(self, other):
        return other == self._model_type


    def __str__(self):
        return self.get_model_name()

    
    def __getstate__(self):
        """
            Called when pickeling the object.
            Remove logger, because no use to pickle logger.
        """
        d = dict(self.__dict__)
        del d["logger"]
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        verbose = d["__verbose"]
        self.logger = setup_logger(verbose, "Model Logger")
